import subprocess


class Info(dict):
    'Dict that acts like an object.'
    def __getattr__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        self[attr] = value


def run(cmd, shell=True, debug=False):
    'Run a command and return the output.'
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=shell)
    (out, _) = proc.communicate()  # no need for stderr
    if debug:
        print cmd
        print out
    return out





'''
    windows.py - (C) 2012-13, Mike Miller
    License: GPLv3+.
'''
import sys, os, locale, stat, platform
try:
    from winstats import ( _diskusage, get_drives, get_drive_type, get_fs_usage,
                           get_perf_data, get_perf_info, get_mem_info,
                           get_vol_info )
except ImportError:
    sys.exit('Error: winstats module not found.  C:\> pip install winstats')

try:
    import colorama
    colorama.init()
    coloravail = True
except ImportError:
    coloravail = False

from fr.utils import Info


version = platform.win32_ver()[1]
win7ver = '6.1.7600'
vistver = '6.0.6000'

hicolor  = False
boldbar  = True
col_lblw = 'CACHE'
col_lbls = 'CACHE'

# icons and graphics chars
encoding = 'cp437'
locale.setlocale(locale.LC_ALL, ('english_united-states', '437'))
locale.setlocale(locale.LC_NUMERIC, '')
_ramico  = '\xf0'       # three horz lines
_diskico = '\xfe'       # block
_unmnico = '\7f'        # empty house
_remvico = '\x1d'       # <->
_netwico = '\x17'       # ^V
_discico = '\xe8'       # cap phi
_emptico = '\xed'       # lc phi
_ellpico = '\xaf'       # >>
_usedico = '\xfe'       # block
_cmonico = '\x04'       # cache mono
_freeico = '\xf9'       # dot
_brckico = ('\xb3', '\xb3') # vert line

_drive_type_result = {
    0: ('unk', None),       # UNKNOWN
    1: ('err', None),       # NO_ROOT_DIR
    2: ('isrem', True),     # REMOVABLE
    3: ('isrem', False),    # FIXED
    4: ('isnet', True),     # REMOTE
    5: ('isopt', True),     # CDROM
    6: ('isram', True),     # RAMDISK
}


def get_diskinfo(outunit, show_all=False, debug=False, local_only=False):
    ''' Returns a list holding the current disk info,
        stats divided by the ouptut unit.
    '''
    disks = {}
    for drive in get_drives():
        drive += ':\\'
        disk = Info(dev=drive)
        try:  usage = get_fs_usage(drive)
        except WindowsError:  # disk not ready, request aborted, etc.
            if show_all:  usage = _diskusage(0, 0, 0)
            else:         continue
        disk.ocap   = usage.total
        disk.cap    = usage.total / outunit
        disk.used   = usage.used / outunit
        disk.free   = usage.free / outunit
        disk.label  = get_vol_info(drive).name
        if usage.total:
            disk.pcnt = float(usage.used) / usage.total * 100
        else:
            disk.pcnt = 0
        disk.mntp   = ''
        disk.ismntd = True
        disk.isnet  = None
        disk.isopt  = None
        disk.isram  = None
        disk.isrem  = None
        dtint, dtstr = get_drive_type(drive)
        setattr(disk, *_drive_type_result[dtint])
        #~ disk.rw = os.access(drive, os.W_OK) # doesn't work on optical
        if usage.total:    # this not giving correct result on Win7 RTM either
            disk.rw = stat.S_IMODE(os.stat(drive).st_mode) & stat.S_IWRITE
        else:
            disk.rw = False
        disks[drive] = disk

    if debug:  print disks
    keys = sorted(disks.keys())
    return [ disks[dev]  for dev in keys ]



def get_meminfo(outunit, debug=False):
    ''' Returns a dictionary holding the current memory info,
        divided by the ouptut unit.
    '''
    meminfo = Info()
    mstat = get_mem_info()  # from winstats
    pinf = get_perf_info()
    try:
        pgpcnt = get_perf_data(r'\Paging File(_Total)\% Usage', 'double')[0] / 100
    except WindowsError:
        pgpcnt = 0

    totl = mstat.TotalPhys
    meminfo.MemTotal = totl / float(outunit)
    used = totl * mstat.MemoryLoad / 100.0  # percent, more reliable
    meminfo.Used = used / float(outunit)
    left = totl - used

    # Cached
    cach = pinf.SystemCacheBytes
    if cach > left and version >= win7ver:
        # Win7 RTM bug :/ this cach number is bogus
        free = get_perf_data(r'\Memory\Free & Zero Page List Bytes', 'long')[0]
        cach = left - free
        meminfo.MemFree = free / float(outunit)
    else:
        meminfo.MemFree = (totl - used - cach) / float(outunit)
    meminfo.Buffers = 0

    meminfo.Cached = cach / float(outunit)

    # SWAP  these numbers are actually commit charge, not swap; fix
    #       should not contain RAM :/
    swpt = abs(mstat.TotalPageFile - totl)
    # these nums aren't quite right either, use perfmon instead :/
    swpu = swpt * pgpcnt
    swpf = swpt - swpu

    meminfo.SwapTotal = swpt / float(outunit)
    meminfo.SwapFree = swpf / float(outunit)
    meminfo.SwapUsed = swpu / float(outunit)
    meminfo.SwapCached = 0  # A linux stat for compat

    if debug:
        import locale
        fmt = lambda x: locale.format('%d', x, True)
        print
        print 'TotalPhys:', fmt(totl)
        print 'AvailPhys:', fmt(mstat.AvailPhys)
        print 'MemoryLoad:', fmt(mstat.MemoryLoad)
        print
        print 'used:', fmt(used)
        print 'left:', fmt(left)
        if 'free' in locals(): print 'PDH Free:', fmt(free)
        print 'SystemCacheBytes:', fmt(pinf.SystemCacheBytes)
        print
        print 'TotalPageFile:', fmt(mstat.TotalPageFile)
        print 'AvailPageFile:', fmt(mstat.AvailPageFile)
        print 'TotalPageFile fixed:', fmt(swpt)
        print 'AvailPageFile fixed:', fmt(swpf)

    return meminfo




'''
    darwin.py - (C) 2013, Jonathan Eunice
    derived from linux.py - (C) 2012-13, Mike Miller
    License: GPLv3+.
'''
import os, locale, re
from os.path import basename
from fr.utils import Info, run

devfilter   = ('none', 'tmpfs', 'udev', 'devfs', 'map')
diskcmd     = '/bin/df -k'
coloravail  = True
hicolor     = None
boldbar     = None
encoding    = 'utf8'
col_lblw    = 'MOUNT CACHE'
col_lbls    = 'MNT CACHE'
TERM        = os.environ.get('TERM')

if TERM == 'xterm-256color':
    hicolor     = True
    boldbar     = False
elif TERM == 'xterm':
    hicolor     = False
    boldbar     = True
else:
    coloravail = False

locale.setlocale(locale.LC_ALL, '')


def get_diskinfo(outunit, show_all=False, debug=False, local_only=False):
    ''' Returns a list holding the current disk info,
        stats divided by the ouptut unit.
    '''
    disks = {}
    try:
        lines = run(diskcmd).splitlines()[1:]  # dump header
        for i, line in enumerate(lines):
            if line:
                tokens  = line.split()
                dev = basename(tokens[0])
                disk = Info()
                disk.isram  = None
                if dev in devfilter:
                    if show_all:
                        if tokens[0] == 'map':
                            tokens = [ ' '.join(tokens[0:2]) ] + tokens[2:]
                            dev = basename(tokens[0])
                        dev += str(i)
                        disk.isram = True
                    else:
                        disk.isram = False; continue
                disk.dev = dev
                # convert to bytes, then output units
                disk.ocap   = float(tokens[1]) * 1024
                disk.cap    = disk.ocap / outunit
                disk.free   = float(tokens[3]) * 1024 / outunit
                disk.label  = '' # dummy values for now
                disk.mntp   = ' '.join( tokens[8:] )
                disk.pcnt   = int(tokens[4][:-1])
                disk.used   = float(tokens[2]) * 1024 / outunit
                disk.ismntd = True
                disk.isnet  = ':' in tokens[0]  # cheesy but works
                if local_only and disk.isnet:
                    continue
                disk.isopt  = None
                disk.isrem  = None
                disk.rw     = True

                disks[dev] = disk
    except IOError:
        return None

    if debug:  print disks
    devices = sorted(disks.keys())
    return [ disks[device]  for device in devices ]


def get_meminfo(outunit, debug=False):
    ''' Returns a dictionary holding the current memory info,
        divided by the ouptut unit.  If mem info can't be read, returns None.
        For Darwin / Mac OS X, interrogates the output of the sysctl and
        vm_stat utilities rather than /proc/meminfo
    '''
    meminfo = Info()
    ms = run('/usr/sbin/sysctl -a hw.memsize'.split(), shell=False)
    su = run('/usr/sbin/sysctl -a vm.swapusage'.split(), shell=False)
    vm = run('/usr/bin/vm_stat', shell=False)

    # Process memsize info info
    memsize = int(ms.split()[1])

    # Process vm_stat
    vmLines = vm.split('\n')
    LINE_RE = re.compile(r'^(.+)\: +(\d+)')
    PAGESIZE = 4096

    vmStats = Info()
    for row in vmLines[1:8]:
        m = LINE_RE.search(row.strip())
        if m:
            name = m.group(1).split()[1]
            vmStats[name] = int(m.group(2)) * PAGESIZE
        else:
            raise ValueError("Can't parse '{0}'".format(row))

    meminfo.MemTotal = memsize / outunit
    meminfo.MemFree  = vmStats.free / outunit
    meminfo.Used     = (vmStats.wired + vmStats.active) / outunit
    meminfo.Cached   = (vmStats.inactive + vmStats.speculative) / outunit
    meminfo.Buffers  = 0

    # Parse vm.swapusage line from sysctl
    SWAPUSAGE_RE = re.compile(
     r'vm\.swapusage: total = ([\d\.]+)M  used = ([\d\.]+)M  free = ([\d\.]+)M'
    )
    m = SWAPUSAGE_RE.search(su.strip())
    if m:
        meminfo.SwapTotal = int(float(m.group(1)))
        meminfo.SwapUsed  = int(float(m.group(2)))
        meminfo.SwapFree  = int(float(m.group(3)))
        meminfo.SwapCached = 0
    else:
        raise ValueError("Can't parse vm.swapusage line from sysctl '{0}'"
                        ).format(su)

    if debug:  print meminfo
    return meminfo




'''
    linux.py - (C) 2012-13, Mike Miller
    License: GPLv3+.
'''
import os, locale
from os.path import basename
from fr.utils import Info, run

try:
    import dbus
except ImportError:
    dbus = None

devfilter   = ('none', 'tmpfs', 'udev', 'cgroup_root')
diskcmd     = '/bin/df --portability'
#~ diskcmd     = '/bin/cat df.txt'  # for testing
memfname    = '/proc/meminfo'
coloravail  = True
hicolor     = None
boldbar     = None
encoding    = 'utf8'
col_lblw    = 'MOUNT CACHE'
col_lbls    = 'MNT CACHE'
TERM        = os.environ.get('TERM')

if TERM == 'linux':  # basic console, use ascii
    _ramico     = 'r'
    _diskico    = 'd'
    _unmnico    = 'u'
    _remvico    = 'm'
    _netwico    = 'n'
    _discico    = 'o'
    _emptico    = '0'
    _ellpico    = '~'
    _usedico    = '#'
    _freeico    = '-'
    _warnico    = '!'
    _brckico    = ('[', ']')
    hicolor     = False
    boldbar     = True
elif TERM == 'xterm-256color':
    hicolor     = True
    boldbar     = False

locale.setlocale(locale.LC_ALL, '')


def get_diskinfo(outunit, show_all=False, debug=False, local_only=False):
    ''' Returns a list holding the current disk info,
        stats divided by the ouptut unit.

        Udisks doesn't provide free/used info, so it is currently gathered
        via the df command.
    '''
    disks = {}
    try:
        lines = run(diskcmd).splitlines()[1:]  # dump header
        for i, line in enumerate(lines):
            if line:
                tokens  = line.split()
                disk = Info()
                disk.isnet  = ':' in tokens[0]  # cheesy but works
                if local_only and disk.isnet:
                    continue
                dev = basename(tokens[0])
                disk.isram  = None
                if dev in devfilter:
                    if show_all:
                        dev += str(i)
                        disk.isram = True
                    else:
                        disk.isram = False; continue
                disk.dev = dev
                # convert to bytes, then output units
                disk.ocap   = float(tokens[1]) * 1024
                disk.cap    = disk.ocap / outunit
                disk.free   = float(tokens[3]) * 1024 / outunit
                disk.label  = '' # dummy values for now
                disk.mntp   = ' '.join( tokens[5:] )
                if disk.mntp == '/boot/efi' and not show_all:
                    continue  # skip
                disk.pcnt   = int(tokens[4][:-1])
                disk.used   = float(tokens[2]) * 1024 / outunit
                disk.ismntd = True
                disk.isopt  = None
                disk.isrem  = None
                disk.rw     = True

                disks[dev] = disk
    except IOError:
        return None

    if dbus:  # request volume details, add to disks
        err = get_udisks2_info(disks, show_all, debug)
        if err:  # fall back to udisks 1
            err = get_udisks_info(disks, show_all, debug)
        if err and 'ServiceUnknown' in str(err):
            print _warnico + ' Udisks not avail.'

    if debug:  print disks
    devices = sorted(disks.keys())
    return [ disks[device]  for device in devices ]


def get_udisks_info(disks, show_all, debug):
    dbus_api = 'org.freedesktop.UDisks'
    dbus_dev = 'org.freedesktop.UDisks.Device'
    try:
        bus = dbus.SystemBus() # Get on the bus... and pay your fare...
        udisks = bus.get_object(dbus_api, '/' + dbus_api.replace('.','/'))
        udisks = dbus.Interface(udisks, dbus_api)

        isnet = False  # only local drives enumerated here
        for devpath in udisks.EnumerateDevices():
            devobj = bus.get_object(dbus_api, devpath)
            devobj = dbus.Interface(devobj, dbus.PROPERTIES_IFACE)

            dev = basename(devpath)
            lbl = unicode(devobj.Get(dbus_dev, 'IdLabel'))
            ismntd = bool(devobj.Get(dbus_dev, 'DeviceIsMounted'))
            ispart = devobj.Get(dbus_dev, 'DeviceIsPartition')
            dtype = devobj.Get(dbus_dev, 'IdType')
            ptype = devobj.Get(dbus_dev, 'PartitionType')
            psize = devobj.Get(dbus_dev, 'PartitionSize')
            isopt = bool(devobj.Get(dbus_dev, 'DeviceIsOpticalDisc'))
            #~ isrem = devobj.Get(dbus_dev, 'DeviceIsRemovable')# !reliable
            isrem = not devobj.Get(dbus_dev, 'DeviceIsSystemInternal')

            rw = False
            if ismntd:
                firstmnt = devobj.Get(dbus_dev, 'DeviceMountPaths')[0]
                if firstmnt == '/':   rw = True
                else:                 rw = os.access(firstmnt, os.W_OK)

                thisone = dict(dev=dev, label=lbl, rw=rw,
                               ismntd=ismntd, isopt=isopt, isrem=isrem,
                               isram=False, isnet=isnet)
                if dev in disks:
                    disks[dev].update(thisone)
                else:
                    disks[dev] = Info(**thisone)

                disks[dev].setdefault('mntp', '')
                disks[dev].setdefault('pcnt', 0)
            else:
                if show_all:
                    if dtype != 'swap' and ptype != '0x05': # extended part.
                        disk = Info(dev=dev, cap=0, used=0, free=0,
                                pcnt=0, ocap=0, mntp='', label=lbl,
                                rw=rw, ismntd=ismntd, isrem=isrem,
                                isopt=isopt, isram=0, isnet=isnet)
                        disks[dev] = disk

            if debug:
                print devpath, dtype, ptype, psize,
                print 'rem:', isrem, ' opt:', isopt,
                print ' part:', ispart, ' rw:', rw, ' mnt:', ismntd
        err = None
    except dbus.exceptions.DBusException as err:
        pass
    return err


def get_udisks2_info(disks, show_all, debug):
    try:
        dbus_api = 'org.freedesktop.UDisks2'
        dbus_base = 'org.freedesktop.UDisks2.'
        dbus_dev_blk = dbus_base + 'Block'
        dbus_dev_drv = dbus_base + 'Drive'
        dbus_dev_fss = dbus_base + 'Filesystem'
        dbus_mgr = 'org.freedesktop.DBus.ObjectManager'

        bus = dbus.SystemBus()
        udisks2 = bus.get_object(dbus_api, '/' + dbus_api.replace('.','/'))
        udisks2 = dbus.Interface(udisks2, dbus_mgr)
        isnet = None  # only local drives enumerated here
        udisk_drives = {}

        # first look at drives, these are in the same list :/
        for devpath, obj in udisks2.GetManagedObjects().iteritems():
            if dbus_dev_drv in obj:
                disk_info = obj.get(dbus_dev_drv, {})
                udisk_drives[devpath] = {
                    'isejt': bool(disk_info['Ejectable']),
                    'isopt': bool(disk_info['Optical']),
                    'isrem': bool(disk_info['Removable']), # Removable
                }

        # get filesystem info
        for devpath, obj in udisks2.GetManagedObjects().iteritems():
            if dbus_dev_fss in obj:
                dev = basename(devpath)
                block_info = obj.get(dbus_dev_blk, {})
                drive_path = str(block_info['Drive'])
                lbl = str(block_info['IdLabel'])
                rw = False

                # get mount point
                ismntd, firstmnt = None, None
                fsys_info = obj.get(dbus_dev_fss, {})
                if fsys_info:
                    mounts = fsys_info.get('MountPoints')
                    if mounts:
                        # convert from array of bytes to string w/o null :/
                        firstmnt = ''.join([ chr(c) for c in mounts[0] if c ])
                        if firstmnt == '/boot/efi' and not show_all:
                            continue  # skip
                        ismntd = True

                if ismntd: # mounted, check read-write
                    if firstmnt == '/':
                        rw = True  # force True
                    else:
                        rw = os.access(firstmnt, os.W_OK)
                    thisone = dict( label=lbl, drive=drive_path,
                                    isnet=isnet,
                                    ismntd=ismntd, firstmnt=firstmnt, rw=rw,
                              )
                    thisone.update(**udisk_drives.get(drive_path, {}))
                    if dev in disks:
                        disks[dev].update(thisone)
                    else:
                        disks[dev] = Info(**thisone)

                    disks[dev].setdefault('mntp', firstmnt or '')
                    disks[dev].setdefault('pcnt', 0)
                else:
                    if show_all:
                        if True:    # dtype != 'swap' and ptype != '0x05':
                                    # extended part.
                            disk = Info(dev=dev, cap=0, used=0, free=0,
                                    pcnt=0, ocap=0, mntp='', label=lbl,
                                    rw=rw, ismntd=ismntd, isram=0, isnet=isnet,
                                    **udisk_drives.get(drive_path, {}))
                            disks[dev] = disk
        err = None
    except dbus.exceptions.DBusException as err:
        pass
    return err


def get_meminfo(outunit, debug=False):
    ''' Returns a dictionary holding the current memory info,
        divided by the ouptut unit.  If mem info can't be read, returns None.
    '''
    meminfo = Info()
    if os.access(memfname, os.R_OK):
        memf = file(memfname, 'r')
    else:
        return None

    for line in memf:  # format: 'MemTotal:  511456 kB\n'
        tokens = line.split()
        if tokens:
            name, value = tokens[0][:-1], tokens[1]  # rm :
            if len(tokens) > 2:
                unit = tokens[2].lower()
            # parse_result to bytes
            value = int(value)
            if   unit == 'kb': value = value * 1024  # most likely
            elif unit ==  'b': value = value
            elif unit == 'mb': value = value * 1024 * 1024
            elif unit == 'gb': value = value * 1024 * 1024 * 1024
            setattr(meminfo, name, value / float(outunit))

    cach = meminfo.Cached + meminfo.Buffers
    meminfo.Used = meminfo.MemTotal - meminfo.MemFree - cach
    meminfo.SwapUsed = (meminfo.SwapTotal - meminfo.SwapCached -
                        meminfo.SwapFree)
    return meminfo




#!/usr/bin/env python
import os
from distutils.core import setup

# grab metadata
version = '1.00'
with open('fr/fr') as f:
    for line in f:
        if line.lstrip().startswith('__version__'):
            try:
                version = line.split("'")[1]
            except IndexError:  pass
            break

# readme is needed at upload time, not install time
try:
    with open('readme.rst') as f:                  # no unicode for older vers.
        long_description = f.read().decode('utf8') #.encode('ascii', 'replace')
except IOError:
    long_description = ''

# install helper script?
scripts = ['fr/fr']
if os.name == 'nt':
    scripts.append('fr.cmd')


setup(
    name          = 'fr',
    version       = version,
    description   = 'A command-line tool to print free resources in' +
                    ' delicious flavors.',
    author        = 'Mike Miller',
    author_email  = 'mixmastamyk@bitbucket.org',
    url           = 'https://bitbucket.org/mixmastamyk/fr',
    download_url  = 'https://bitbucket.org/mixmastamyk/fr/get/default.tar.gz',
    license       = 'GPLv3',
    packages      = ['fr'],
    scripts       = scripts,
    extras_require = {
        'win': ['winstats'],
        'color': ['colorama'],
    },

    long_description = long_description,
    classifiers     = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: Microsoft :: Windows :: Windows XP',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: System :: Hardware',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
)



'''
    ansilib, a library of common console color functions.
    (C) 2005-12, Mike Miller - Released under the GPL, version 3+.
'''
import sys, os

if True:  # fold init
    # fg
    black       = 30
    red         = 31
    green       = 32
    yellow      = 33
    orange      = 33
    blue        = 34
    purple      = 35
    cyan        = 36
    grey        = 37

    # bg
    blackbg     = 40
    redbg       = 41
    greenbg     = 42
    orangebg    = 43
    bluebg      = 44
    purplebg    = 45
    cyanbg      = 46
    greybg      = 47

    # attrs
    norm        = 0
    bold        = 1
    dim         = 2
    underline   = 4
    reverse     = 7
    strike      = 9
    dimbb       = '%s;%s' % (bold, black) # works in more places


    # A list of ansi escape sequences.
    fred        = '[00;31m%s[00m'
    fbred       = '[01;31m%s[00m'
    fgreen      = '[00;32m%s[00m'
    fbgreen     = '[01;32m%s[00m'
    forange     = '[00;33m%s[00m'
    fbyellow    = '[01;33m%s[00m'
    fblue       = '[00;34m%s[00m'
    fbblue      = '[01;34m%s[00m'
    fpurple     = '[00;35m%s[00m'
    fbpurple    = '[01;35m%s[00m'
    fcyan       = '[00;36m%s[00m'
    fbcyan      = '[01;36m%s[00m'
    fgrey       = '[00;37m%s[00m'
    fwhite      = '[01;37m%s[00m'
    #fgrey       = '[00;38m%s[00m'
    #fwhite      = '[01;38m%s[00m'

    redrev      = '[00;05;37;41m%s[00m'
    grerev      = '[00;05;37;42m%s[00m'
    yelrev      = '[01;05;37;43m%s[00m'

    rev         = '[07m%s[00m'
    fbold       = '[01m%s[00m'
    fdim        = '[02m%s[00m'
    fdimbb      = '[1;30m%s[0m'

    # Readline encoded escape sequences:
    greenprompt  = '\001[01;32m\002%s\001[00m\002'
    yellowprompt = '\001[01;33m\002%s\001[00m\002'
    redprompt    = '\001[01;31m\002%s\001[00m\002'

    # 256...
    darkred = 52

    # 256 color support
    csi8 = '\x1b[0;38;5;%03dm'
    csi8b = '\x1b[1;38;5;%03dm'
    csi8_blk = csi8b + '%%s\x1b[0m'
    grn8 = 34
    grnyl8 = 76
    yell8 = 184
    orng8 = 172
    orngrd8 = 166
    red8 = 160
    blu8 = 27
    pal8 = [grn8, grn8, grn8, grn8, grn8, grnyl8, yell8, orng8, orngrd8, red8]
    map8 = { blue: blu8, red: red8, green: grn8, dim: dim }
    rst8 = grn8

    # 16 colors
    csi4 = '\x1b[0;%sm'
    csi4b = '\x1b[1;%sm'
    csi4_blk = csi4b + '%%s\x1b[0m'
    yel = yellow
    pal4 = [green, green, green, green, green, yel, yel, yel, yel, red]
    rst4 = green


def colorstart(fgcolor, bgcolor, weight):
    '''Begin a text style.'''
    if weight:          weight = bold
    else:               weight = norm
    if bgcolor:
        sys.stdout.write('[%s;%s;%sm' % (weight, fgcolor, bgcolor))
    else:
        sys.stdout.write('[%s;%sm' % (weight, fgcolor))


def colorend(cr=False):
    '''End color styles.  Resets to default terminal colors.'''
    sys.stdout.write('[0m')
    if cr: sys.stdout.write('\n')
    sys.stdout.flush()


def cprint(text, fg=grey, bg=blackbg, w=norm, cr=False, encoding='utf8'):
    ''' Print a string in a specified color style and then return to normal.
        def cprint(text, fg=white, bg=blackbg, w=norm, cr=True):
    '''
    colorstart(fg, bg, w)
    sys.stdout.write(text)
    colorend(cr)


def bargraph(data, maxwidth, incolor=True, cbrackets=(u'\u2595', u'\u258F')):
    'Creates a bar graph.'
    threshold = 100.0 / (maxwidth * 2)  # if smaller than 1/2 of one char wide
    position = 0
    begpcnt = data[0][1] * 100
    endpcnt = data[-1][1] * 100

    if len(data) < 1: return        # Nada to do
    maxwidth = maxwidth - 2         # because of brackets
    datalen = len(data)

    # Print left bracket in correct color:
    if cbrackets and incolor: # and not (begpcnt == 0 and endpcnt == 0):
        if begpcnt < threshold: bkcolor = data[-1][2]  # greenbg
        else:                   bkcolor = data[0][2]   # redbg
        cprint(cbrackets[0], data[0][2], bkcolor, None, None)
    else:
        sys.stdout.write(cbrackets[0])

    for i, part in enumerate(data):
        # unpack data
        char, pcnt, fgcolor, bgcolor, bold = part
        width = int( round(pcnt/100.0 * maxwidth, 0) )
        position = position + width

        # and graph
        if incolor and not ( fgcolor is None):
            cprint(char * width, fgcolor, bgcolor, bold, False)
        else:
            sys.stdout.write((char * width))

        if i == (datalen - 1):   # correct last one
            if position < maxwidth:
                if incolor: # char
                    cprint(char * (maxwidth-position), fgcolor, bgcolor,
                           bold, False)
                else:
                    sys.stdout.write(char * (maxwidth-position))
            elif position > maxwidth:
                sys.stdout.write(chr(8) + ' ' + chr(8))  # backspace

    # Print right bracket in correct color:
    if cbrackets and incolor:
        if endpcnt < threshold: bkcolor = data[0][3] # redbg
        else:                   bkcolor = data[1][3] # greenbg
        cprint(cbrackets[1], data[-1][2], bkcolor, None, None)
    else:
        sys.stdout.write(cbrackets[1])


def get_palette(hicolor):
    if hicolor:
        return csi8, csi8b, csi8_blk, pal8, rst8, len(pal8)
    else:
        return csi4, csi4b, csi4_blk, pal4, rst4, len(pal4)


def get_label_tmpl(value, maxwidth, hicolor):
    csi, csib, blk, pal, rst, plen = get_palette(hicolor)
    # not happy with calculating this again...
    pos = int(maxwidth * (value / 100.0)) - 1
    colorind = get_color_index(pos, 0, maxwidth, plen)
    return blk % pal[colorind]


def get_color_index(pos, offset, maxwidth, plen):
    bucket = float(maxwidth) / plen
    return min(int((pos+offset)/bucket), (plen-1))


def rainbar(data, maxwidth, incolor=True, hicolor=True,
            cbrackets=(u'\u2595', u'\u258F')):
    'Creates a "rainbar" graph.'
    if not data: return             # Nada to do
    datalen = len(data)
    endpcnt = data[-1][1] # * 100
    maxwidth = maxwidth - 2         # because of brackets

    # setup
    out = sys.stdout
    csi, csib, _, pal, rst, plen = get_palette(hicolor)

    empty = data[-1][0]
    bucket = float(maxwidth) / plen
    position = 0

    # Print left bracket in correct color:
    if cbrackets and incolor:
        out.write((csi % pal[0]) + cbrackets[0])  # start bracket
    else:
        out.write(cbrackets[0])

    for i, part in enumerate(data):
        char, pcnt, fgcolor, bgcolor, bold = part
        if fgcolor and hicolor:
            fgcolor = map8[fgcolor]
        if not bold:
            csib = csi

        lastind = None
        width = int(maxwidth * (pcnt / 100.0))
        offset = position
        position += width

        if incolor:
            for j in range(width):
                # faster:
                colorind = fgcolor or min(int((j+offset)/bucket), (plen-1))
                #~ colorind = fgcolor or get_color_index(j, offset, maxwidth, plen)
                if colorind == lastind:
                    out.write(char)
                else:
                    color = fgcolor or pal[colorind]
                    out.write((csib % color) + char)
                lastind = colorind
        else:
            out.write((char * width))

        if i == (datalen - 1):          # check if last one correct
            if position < maxwidth:
                rest = maxwidth - position
                if incolor: # char
                    out.write((csib % pal[-1]) + (empty * rest))
                else:
                    out.write(char * rest)
            elif position > maxwidth:
                out.write(chr(8) + ' ' + chr(8))  # backspace

    # Print right bracket in correct color:
    if cbrackets and incolor:
        lastcolor = darkred if (hicolor and endpcnt > 1) else pal[-1]
        out.write((csi % lastcolor) + cbrackets[1])    # end bracket
        colorend()
    else:
        out.write(cbrackets[1])


# -------------------------------------------------------------------
# modified from gentoo portage "output" module
# Copyright 1998-2004 Gentoo Foundation
def set_xterm_title(mystr):
    'set the title of an xterm'
    if os.environ.has_key('TERM') and sys.stderr.isatty():
        terms = ['xterm', 'Eterm', 'aterm', 'rxvt', 'screen', 'kterm',
                 'rxvt-unicode']
        term = os.environ['TERM']
        if term in terms:
            sys.stderr.write('\x1b]2;'+str(mystr)+'\x07')
            sys.stderr.flush()

# -------------------------------------------------------------------
# from avkutil.py - Andrei Kulakov <ak@silmarill.org>
def get_term_size():
    '''Return terminal size as tuple (width, height).'''
    x, y = 0, 0
    if sys.platform.startswith('win'):
        #~ http://code.activestate.com/recipes/440694-determine-size-of-console-window-on-windows/
        res = None
        try:
            from ctypes import windll, create_string_buffer
            stderr = -12
            h = windll.kernel32.GetStdHandle(stderr)
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
        except Exception as err:
            print 'ERROR:', err.__class__.__name__, str(err)
        if res:
            import struct
            (bufx, bufy, curx, cury, wattr, left, top, right, bottom,
             maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
            x = right - left + 1
            y = bottom - top + 1
    else:  # Mac, Linux
        import struct, fcntl, termios
        try:
            y, x = struct.unpack('hhhh', fcntl.ioctl(0,
                termios.TIOCGWINSZ, '\000'*8))[0:2]
        except IOError:
            pass
    return x, y

